# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect

import psutil
import rclpy

from ros2cli.node.direct import DirectNode


def get_interfaces_ip_addresses():
    addresses_by_interfaces = psutil.net_if_addrs()
    print(f'Addresses by interfaces: {addresses_by_interfaces}')
    return addresses_by_interfaces


class NetworkAwareNode:
    """A direct node, that resets itself when a network interface changes."""

    def __init__(self, args):
        self.args = args
        # TODO(ivanpauno): A race condition is possible here, since it isn't possible to know
        # exactly which interfaces were available at node creation.
        self.node = DirectNode(args)
        self.addresses_at_start = get_interfaces_ip_addresses()

    def __enter__(self):
        self.node.__enter__()
        return self

    def __getattr__(self, name):
        attr = getattr(self.node, name)

        if inspect.ismethod(attr):
            @functools.wraps(attr)
            def wrapper(*args, **kwargs):
                self.reset_if_addresses_changed()
                # The attribute has to be get here again, in case self.node changed
                return getattr(self.node, name)(*args, **kwargs)
            wrapper.__signature__ = inspect.signature(attr)
            return wrapper
        self.reset_if_addresses_changed()
        return attr

    def __exit__(self, exc_type, exc_value, traceback):
        self.node.__exit__(exc_type, exc_value, traceback)

    def reset_if_addresses_changed(self):
        new_addresses = get_interfaces_ip_addresses()
        if new_addresses != self.addresses_at_start:
            self.addresses_at_start = new_addresses
            self.node.destroy_node()
            rclpy.shutdown()
            self.node = DirectNode(self.args)
            self.node.__enter__()
            print('Network interfaces changed, daemon node was reset!')



# Copyright 2016-2017 Dirk Thomas
# Copyright 2017 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import logging

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

"""
The group name for entry points identifying extension points.

While all entry points in this package start with ``ros2cli.`` other
distributions might define entry points with a different prefix.
Those need to be declared using this group name.
"""
EXTENSION_POINT_GROUP_NAME = 'ros2cli.extension_point'

logger = logging.getLogger(__name__)


def get_all_entry_points():
    """
    Get all entry points related to ``ros2cli`` and any of its extensions.

    :returns: mapping of entry point names to ``EntryPoint`` instances
    :rtype: dict
    """
    extension_points = get_entry_points(EXTENSION_POINT_GROUP_NAME)

    entry_points = defaultdict(dict)

    for dist in importlib_metadata.distributions():
        for ep in dist.entry_points:
            # skip groups which are not registered as extension points
            if ep.group not in extension_points:
                continue

            entry_points[ep.group][ep.name] = (dist, ep)
    return entry_points


def get_entry_points(group_name):
    """
    Get the entry points for a specific group.

    :param str group_name: the name of the ``entry_point`` group
    :returns: mapping of group name to dictionaries which map entry point names
      to ``EntryPoint`` instances
    :rtype: dict
    """
    entry_points_impl = importlib_metadata.entry_points()
    if hasattr(entry_points_impl, 'select'):
        groups = entry_points_impl.select(group=group_name)
    else:
        groups = entry_points_impl.get(group_name, [])
    entry_points = {}
    for entry_point in groups:
        entry_points[entry_point.name] = entry_point
    return entry_points


def load_entry_points(group_name, *, exclude_names=None):
    """
    Load the entry points for a specific group.

    :param str group_name: the name of the ``entry_point`` group
    :param iterable exclude_names: the names of the entry points to exclude
    :returns: mapping of entry point name to loaded entry point
    :rtype: dict
    """
    extension_types = {}
    for entry_point in get_entry_points(group_name).values():
        if exclude_names and entry_point.name in exclude_names:
            continue
        try:
            extension_type = entry_point.load()
        except Exception as e:  # noqa: F841
            logger.warning(
                f"Failed to load entry point '{entry_point.name}': {e}")
            continue
        extension_types[entry_point.name] = extension_type
    return extension_types


def get_first_line_doc(any_type):
    if not any_type.__doc__:
        return ''
    lines = any_type.__doc__.splitlines()
    if not lines:
        return ''
    if lines[0]:
        line = lines[0]
    elif len(lines) > 1:
        line = lines[1]
    return line.strip().rstrip('.')



# Copyright 2021 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copyreg
import os
import pickle
import platform
import socket
import subprocess
import sys
import threading

from ros2cli.helpers import wait_for


class PicklerForProcess(pickle.Pickler):
    """
    Pickle objects for subprocess.

    This `pickle.Pickler` subclass sends serialized objects
    to the given process through its `stdin` pipe. It can
    serialize and send inheritable sockets and file descriptors.
    """

    def __init__(self, process, *args, **kwargs):
        """
        See `pickle.Pickler` class for further reference.

        :param process: a `subprocess.Popen` instance.
          It is assumed its `stdin` attribute is an open pipe.
        """
        super().__init__(process.stdin, *args, **kwargs)
        self.process = process
        self.dispatch_table = copyreg.dispatch_table.copy()
        self.dispatch_table[socket.socket] = self.reduce_socket
        self.dispatch_table[threading.Event] = self.reduce_event

    def reduce_event(self, obj):
        return threading.Event, ()

    @staticmethod
    def load_socket(data):
        if platform.system() == 'Windows':
            return socket.fromshare(data)
        return socket.socket(fileno=data)

    def reduce_socket(self, obj):
        if platform.system() == 'Windows':
            data = obj.share(self.process.pid)
        else:
            data = obj.fileno()
        return PicklerForProcess.load_socket, (data,)

    def dump(self, *args, **kwargs):
        super().dump(*args, **kwargs)
        self.process.stdin.flush()


def main():
    """
    Execute incoming, serialized callable.

    See `daemonize()` for further reference.
    """
    callable_ = pickle.load(sys.stdin.buffer)
    sys.stdin.close()
    os.close(0)  # force C stream close
    return callable_()


def daemonize(callable_, tags={}, timeout=None, debug=False):
    """
    Spawn a callable object as a daemon.

    The callable object is pickled and sent for execution
    via the `stdin` pipe, which is closed immediately after
    deserialization succeeds. The callable object may hold
    inheritable sockets and file descriptors.

    :param callable_: callable object to be daemonized.
    :param tags: optional key-value pairs to show up as
      command-line '--key value' arguments of the daemon
      process. Useful for identification in the OS process
      list.
    :param timeout: optional duration, in seconds, to wait
      for the daemon to be ready. Non-positive durations will
      result in an indefinite wait.
    :param debug: if `True`, the daemon process will not be
      detached and share both `stdout` and `stderr` streams
      with its parent process.
    """
    # Use daemon `main()` function
    prog = f'from {__name__} import main; main()'
    cmd = [sys.executable, '-c', prog]
    for name, value in tags.items():
        flag = '--' + name.replace('_', '-')
        cmd += [flag, str(value)]
    kwargs = {}
    if platform.system() == 'Windows':
        if not debug:
            kwargs.update(creationflags=subprocess.DETACHED_PROCESS)
        # Avoid showing cmd windows for subprocess
        si = subprocess.STARTUPINFO()
        si.dwFlags = subprocess.STARTF_USESHOWWINDOW
        si.wShowWindow = subprocess.SW_HIDE
        kwargs['startupinfo'] = si
        # Don't keep handle of current working directory in daemon process
        kwargs.update(cwd=os.environ.get('SYSTEMROOT', None))

    kwargs['stdin'] = subprocess.PIPE
    if not debug:
        kwargs['stdout'] = subprocess.DEVNULL
        kwargs['stderr'] = subprocess.DEVNULL
    # Don't close inheritable file descriptors,
    # the given callable object may be carrying
    # some.
    kwargs['close_fds'] = False

    # Spawn child process
    process = subprocess.Popen(cmd, **kwargs)

    # Send serialized callable object through stdin pipe
    pickler = PicklerForProcess(process)
    pickler.dump(callable_)

    if timeout is not None:
        # Wait for daemon to be ready by
        # monitoring when the stdin pipe
        # is closed
        def daemon_ready():
            try:
                pickler.dump(None)
                return False
            except OSError:
                return True
        if not wait_for(daemon_ready, timeout):
            process.terminate()
            raise RuntimeError(
                'Timed out waiting for '
                'daemon to become ready'
            )
    # Make sure the daemon process is still alive
    # (and that the stdin pipe was purposefully closed)
    # before returning.
    if process.poll() is not None:
        rc = process.returncode
        raise RuntimeError(
            'Daemon process died '
            f'with returncode {rc}'
        )
    return
